"""Class definitions for corpora"""
from __future__ import annotations

import logging
import multiprocessing as mp
import os
import random
import subprocess
import sys
import time
from collections import Counter
from queue import Empty
from typing import TYPE_CHECKING, Dict, Generator, List, Optional, Union

import yaml

from ..config import FeatureConfig
from ..exceptions import CorpusError, KaldiProcessingError, TextGridParseError, TextParseError
from ..helper import output_mapping
from ..multiprocessing.classes import Job
from ..multiprocessing.corpus import CorpusProcessWorker
from ..multiprocessing.features import calc_cmvn, compute_vad, mfcc
from ..multiprocessing.helper import Stopped
from ..utils import log_kaldi_errors, thirdparty_binary
from .classes import File, Speaker, Utterance, parse_file
from .helper import find_exts

if TYPE_CHECKING:
    from logging import Logger

    from ..dictionary import DictionaryType


__all__ = ["Corpus"]


class Corpus:
    """
    Class that stores information about the dataset to align.

    Corpus objects have a number of mappings from either utterances or speakers
    to various properties, and mappings between utterances and speakers.

    See http://kaldi-asr.org/doc/data_prep.html for more information about
    the files that are created by this class.


    Parameters
    ----------
    directory : str
        Directory of the dataset to align
    output_directory : str
        Directory to store generated data for the Kaldi binaries
    speaker_characters : int, optional
        Number of characters in the filenames to count as the speaker ID,
        if not specified, speaker IDs are generated from directory names
    num_jobs : int, optional
        Number of processes to use, defaults to 3
    sample_rate : int, optional
        Default sample rate to use for feature generation, defaults to 16000
    debug : bool
        Flag to enable debug mode, defaults to False
    logger : :class:`~logging.Logger`
        Logger to use
    use_mp : bool
        Flag to enable multiprocessing, defaults to True
    punctuation : str, optional
        Characters to treat as punctuation in parsing text
    clitic_markers : str, optional
        Characters to treat as clitic markers in parsing text
    audio_directory : str, optional
        Additional directory to parse for audio files
    skip_load : bool
        Flag to skip loading when initializing, defaults to False
    parse_text_only_files : bool
        Flag to parse text files that do not have associated sound files, defaults to False
    """

    def __init__(
        self,
        directory: str,
        output_directory: str,
        speaker_characters: Union[int, str] = 0,
        num_jobs: int = 3,
        sample_rate: int = 16000,
        debug: bool = False,
        logger: Optional[Logger] = None,
        use_mp: bool = True,
        punctuation: str = None,
        clitic_markers: str = None,
        audio_directory: Optional[str] = None,
        skip_load: bool = False,
        parse_text_only_files: bool = False,
        ignore_speakers: bool = False,
    ):
        self.audio_directory = audio_directory
        self.punctuation = punctuation
        self.clitic_markers = clitic_markers
        self.debug = debug
        self.use_mp = use_mp
        log_dir = os.path.join(output_directory, "logging")
        os.makedirs(log_dir, exist_ok=True)
        self.name = os.path.basename(directory)
        self.log_file = os.path.join(log_dir, "corpus.log")
        if logger is None:
            self.logger = logging.getLogger("corpus_setup")
            self.logger.setLevel(logging.INFO)
            handler = logging.FileHandler(self.log_file, "w", "utf-8")
            handler.setFormatter = logging.Formatter("%(name)s %(message)s")
            self.logger.addHandler(handler)
        else:
            self.logger = logger
        if not os.path.exists(directory):
            raise CorpusError(f"The directory '{directory}' does not exist.")
        if not os.path.isdir(directory):
            raise CorpusError(
                f"The specified path for the corpus ({directory}) is not a directory."
            )

        num_jobs = max(num_jobs, 1)
        if num_jobs == 1:
            self.use_mp = False
        self.original_num_jobs = num_jobs
        self.logger.info("Setting up corpus information...")
        self.directory = directory
        self.output_directory = os.path.join(output_directory, "corpus_data")
        self.temp_directory = os.path.join(self.output_directory, "temp")
        os.makedirs(self.temp_directory, exist_ok=True)
        self.speaker_characters = speaker_characters
        if speaker_characters == 0:
            self.speaker_directories = True
        else:
            self.speaker_directories = False
        self.num_jobs = num_jobs
        self.speakers: Dict[str, Speaker] = {}
        self.files: Dict[str, File] = {}
        self.utterances: Dict[str, Utterance] = {}
        self.sound_file_errors = []
        self.decode_error_files = []
        self.transcriptions_without_wavs = []
        self.no_transcription_files = []
        self.textgrid_read_errors = {}
        self.groups = []
        self.speaker_groups = []
        self.word_counts = Counter()
        self.sample_rate = sample_rate
        if self.use_mp:
            self.stopped = Stopped()
        else:
            self.stopped = False

        self.skip_load = skip_load
        self.utterances_time_sorted = False
        self.parse_text_only_files = parse_text_only_files
        self.feature_config = FeatureConfig()
        self.vad_config = {"energy_threshold": 5.5, "energy_mean_scale": 0.5}
        self.no_speakers = ignore_speakers
        self.vad_segments = {}
        if self.use_mp:
            self.stopped = Stopped()
        else:
            self.stopped = False
        if not self.skip_load:
            self.load()

    def normalized_text_iter(
        self, dictionary: Optional[DictionaryType] = None, min_count: int = 1
    ) -> Generator:
        """
        Construct an iterator over the normalized texts in the corpus

        Parameters
        ----------
        dictionary: :class:`~montreal_forced_aligner.dictionary.Dictionary`
            Dictionary to use for normalization
        min_count: int
            Minimum word count to include in the output, otherwise will use OOV code, defaults to 1

        Yields
        -------
        str
            Normalized text
        """
        unk_words = {k for k, v in self.word_counts.items() if v <= min_count}
        for u in self.utterances.values():
            text = u.text.split()
            new_text = []
            for t in text:
                if dictionary is not None:
                    dictionary.to_int(t)
                    lookup = dictionary.split_clitics(t)
                    if lookup is None:
                        continue
                else:
                    lookup = [t]
                for item in lookup:
                    if item in unk_words:
                        new_text.append("<unk>")
                    elif dictionary is not None and item not in dictionary.words:
                        new_text.append("<unk>")
                    else:
                        new_text.append(item)
            yield " ".join(new_text)

    def subset_directory(self, subset: Optional[int]) -> str:
        """
        Construct a subset directory for the corpus

        Parameters
        ----------
        subset: int, optional
            Number of utterances to include, if larger than the total number of utterance or not specified, the
            split_directory is returned

        Returns
        -------
        str
            Path to subset directory
        """
        if subset is None or subset > self.num_utterances or subset <= 0:
            for j in self.jobs:
                j.set_subset(None)
            return self.split_directory
        directory = os.path.join(self.output_directory, f"subset_{subset}")
        self.create_subset(subset)
        return directory

    def initialize_utt_fsts(self) -> None:
        """
        Construct utterance FSTs
        """
        for j in self.jobs:
            j.output_utt_fsts(self)

    def create_subset(self, subset: Optional[int]) -> None:
        """
        Create a subset of utterances to use for training

        Parameters
        ----------
        subset: int
            Number of utterances to include in subset
        """
        subset_directory = os.path.join(self.output_directory, f"subset_{subset}")

        larger_subset_num = subset * 10
        if larger_subset_num < self.num_utterances:
            # Get all shorter utterances that are not one word long
            utts = sorted(
                (utt for utt in self.utterances.values() if " " in utt.text),
                key=lambda x: x.duration,
            )
            larger_subset = utts[:larger_subset_num]
        else:
            larger_subset = self.utterances.values()
        random.seed(1234)  # make it deterministic sampling
        subset_utts = set(random.sample(larger_subset, subset))
        log_dir = os.path.join(subset_directory, "log")
        os.makedirs(log_dir, exist_ok=True)

        for j in self.jobs:
            j.set_subset(subset_utts)
            j.output_to_directory(subset_directory)

    def load(self) -> None:
        """
        Load the corpus
        """
        loaded = self._load_from_temp()
        if not loaded:
            if self.use_mp:
                self.logger.debug("Loading from source with multiprocessing")
                self._load_from_source_mp()
            else:
                self.logger.debug("Loading from source without multiprocessing")
                self._load_from_source()
        else:
            self.logger.debug("Successfully loaded from temporary files")

    @property
    def file_speaker_mapping(self) -> Dict[str, List[str]]:
        """Speaker ordering for each file"""
        return {file_name: file.speaker_ordering for file_name, file in self.files.items()}

    def _load_from_temp(self) -> bool:
        """
        Load a corpus from saved data in the temporary directory

        Returns
        -------
        bool
            Whether loading from temporary files was successful
        """
        begin_time = time.time()
        for f in os.listdir(self.output_directory):
            if f.startswith("split"):
                old_num_jobs = int(f.replace("split", ""))
                if old_num_jobs != self.num_jobs:
                    self.logger.info(
                        f"Found old run with {old_num_jobs} rather than the current {self.num_jobs}, "
                        f"setting to {old_num_jobs}.  If you would like to use {self.num_jobs}, re-run the command "
                        f"with --clean."
                    )
                    self.num_jobs = old_num_jobs
        speakers_path = os.path.join(self.output_directory, "speakers.yaml")
        files_path = os.path.join(self.output_directory, "files.yaml")
        utterances_path = os.path.join(self.output_directory, "utterances.yaml")

        if not os.path.exists(speakers_path):
            self.logger.debug(f"Could not find {speakers_path}, cannot load from temp")
            return False
        if not os.path.exists(files_path):
            self.logger.debug(f"Could not find {files_path}, cannot load from temp")
            return False
        if not os.path.exists(utterances_path):
            self.logger.debug(f"Could not find {utterances_path}, cannot load from temp")
            return False
        self.logger.debug("Loading from temporary files...")

        with open(speakers_path, "r", encoding="utf8") as f:
            speaker_data = yaml.safe_load(f)

        for entry in speaker_data:
            self.speakers[entry["name"]] = Speaker(entry["name"])
            self.speakers[entry["name"]].cmvn = entry["cmvn"]

        with open(files_path, "r", encoding="utf8") as f:
            files_data = yaml.safe_load(f)
        for entry in files_data:
            self.files[entry["name"]] = File(
                entry["wav_path"], entry["text_path"], entry["relative_path"]
            )
            self.files[entry["name"]].speaker_ordering = [
                self.speakers[x] for x in entry["speaker_ordering"]
            ]
            self.files[entry["name"]].wav_info = entry["wav_info"]

        with open(utterances_path, "r", encoding="utf8") as f:
            utterances_data = yaml.safe_load(f)
        for entry in utterances_data:
            s = self.speakers[entry["speaker"]]
            f = self.files[entry["file"]]
            u = Utterance(
                s,
                f,
                begin=entry["begin"],
                end=entry["end"],
                channel=entry["channel"],
                text=entry["text"],
            )
            self.utterances[u.name] = u
            if u.text:
                self.word_counts.update(u.text.split())
            self.utterances[u.name].features = entry["features"]
            self.utterances[u.name].ignored = entry["ignored"]

        self.logger.debug(
            f"Loaded from corpus_data temp directory in {time.time()-begin_time} seconds"
        )
        return True

    def _load_from_source_mp(self) -> None:
        """
        Load a corpus using multiprocessing
        """
        if self.stopped is None:
            self.stopped = Stopped()
        begin_time = time.time()
        manager = mp.Manager()
        job_queue = manager.Queue()
        return_queue = manager.Queue()
        return_dict = manager.dict()
        return_dict["sound_file_errors"] = manager.list()
        return_dict["decode_error_files"] = manager.list()
        return_dict["textgrid_read_errors"] = manager.dict()
        finished_adding = Stopped()
        procs = []
        for _ in range(self.num_jobs):
            p = CorpusProcessWorker(
                job_queue, return_dict, return_queue, self.stopped, finished_adding
            )
            procs.append(p)
            p.start()
        try:

            use_audio_directory = False
            all_sound_files = {}
            if self.audio_directory and os.path.exists(self.audio_directory):
                use_audio_directory = True
                for root, _, files in os.walk(self.audio_directory, followlinks=True):
                    (
                        identifiers,
                        wav_files,
                        lab_files,
                        textgrid_files,
                        other_audio_files,
                    ) = find_exts(files)
                    wav_files = {k: os.path.join(root, v) for k, v in wav_files.items()}
                    other_audio_files = {
                        k: os.path.join(root, v) for k, v in other_audio_files.items()
                    }
                    all_sound_files.update(other_audio_files)
                    all_sound_files.update(wav_files)

            for root, _, files in os.walk(self.directory, followlinks=True):
                identifiers, wav_files, lab_files, textgrid_files, other_audio_files = find_exts(
                    files
                )
                relative_path = root.replace(self.directory, "").lstrip("/").lstrip("\\")

                if self.stopped.stop_check():
                    break
                if not use_audio_directory:
                    all_sound_files = {}
                    wav_files = {k: os.path.join(root, v) for k, v in wav_files.items()}
                    other_audio_files = {
                        k: os.path.join(root, v) for k, v in other_audio_files.items()
                    }
                    all_sound_files.update(other_audio_files)
                    all_sound_files.update(wav_files)
                for file_name in identifiers:
                    if self.stopped.stop_check():
                        break
                    wav_path = None
                    transcription_path = None
                    if file_name in all_sound_files:
                        wav_path = all_sound_files[file_name]
                    if file_name in lab_files:
                        lab_name = lab_files[file_name]
                        transcription_path = os.path.join(root, lab_name)

                    elif file_name in textgrid_files:
                        tg_name = textgrid_files[file_name]
                        transcription_path = os.path.join(root, tg_name)
                    if wav_path is None and not self.parse_text_only_files:
                        self.transcriptions_without_wavs.append(transcription_path)
                        continue
                    if transcription_path is None:
                        self.no_transcription_files.append(wav_path)
                    job_queue.put(
                        (
                            file_name,
                            wav_path,
                            transcription_path,
                            relative_path,
                            self.speaker_characters,
                            self.sample_rate,
                            self.punctuation,
                            self.clitic_markers,
                        )
                    )

            finished_adding.stop()
            self.logger.debug("Finished adding jobs!")
            job_queue.join()

            self.logger.debug("Waiting for workers to finish...")
            for p in procs:
                p.join()

            while True:
                try:
                    file = return_queue.get(timeout=1)
                    if self.stopped.stop_check():
                        continue
                except Empty:
                    break

                self.add_file(file)

            if "error" in return_dict:
                raise return_dict["error"][1]

            for k in ["sound_file_errors", "decode_error_files", "textgrid_read_errors"]:
                if hasattr(self, k):
                    if return_dict[k]:
                        self.logger.info(
                            "There were some issues with files in the corpus. "
                            "Please look at the log file or run the validator for more information."
                        )
                        self.logger.debug(f"{k} showed {len(return_dict[k])} errors:")
                        if k == "textgrid_read_errors":
                            getattr(self, k).update(return_dict[k])
                            for f, e in return_dict[k].items():
                                self.logger.debug(f"{f}: {e.error}")
                        else:
                            self.logger.debug(", ".join(return_dict[k]))
                            setattr(self, k, return_dict[k])

        except KeyboardInterrupt:
            self.logger.info(
                "Detected ctrl-c, please wait a moment while we clean everything up..."
            )
            self.stopped.stop()
            finished_adding.stop()
            job_queue.join()
            self.stopped.set_sigint_source()
            while True:
                try:
                    _ = return_queue.get(timeout=1)
                    if self.stopped.stop_check():
                        continue
                except Empty:
                    break
        finally:

            if self.stopped.stop_check():
                self.logger.info(f"Stopped parsing early ({time.time() - begin_time} seconds)")
                if self.stopped.source():
                    sys.exit(0)
            else:
                self.logger.debug(
                    f"Parsed corpus directory with {self.num_jobs} jobs in {time.time() - begin_time} seconds"
                )

    def _load_from_source(self) -> None:
        """
        Load a corpus without using multiprocessing
        """
        begin_time = time.time()
        self.stopped = False

        all_sound_files = {}
        use_audio_directory = False
        if self.audio_directory and os.path.exists(self.audio_directory):
            use_audio_directory = True
            for root, _, files in os.walk(self.audio_directory, followlinks=True):
                if self.stopped:
                    return
                identifiers, wav_files, lab_files, textgrid_files, other_audio_files = find_exts(
                    files
                )
                wav_files = {k: os.path.join(root, v) for k, v in wav_files.items()}
                other_audio_files = {
                    k: os.path.join(root, v) for k, v in other_audio_files.items()
                }
                all_sound_files.update(other_audio_files)
                all_sound_files.update(wav_files)

        for root, _, files in os.walk(self.directory, followlinks=True):
            identifiers, wav_files, lab_files, textgrid_files, other_audio_files = find_exts(files)
            relative_path = root.replace(self.directory, "").lstrip("/").lstrip("\\")
            if self.stopped:
                return
            if not use_audio_directory:
                all_sound_files = {}
                wav_files = {k: os.path.join(root, v) for k, v in wav_files.items()}
                other_audio_files = {
                    k: os.path.join(root, v) for k, v in other_audio_files.items()
                }
                all_sound_files.update(other_audio_files)
                all_sound_files.update(wav_files)
            for file_name in identifiers:

                wav_path = None
                transcription_path = None
                if file_name in all_sound_files:
                    wav_path = all_sound_files[file_name]
                if file_name in lab_files:
                    lab_name = lab_files[file_name]
                    transcription_path = os.path.join(root, lab_name)
                elif file_name in textgrid_files:
                    tg_name = textgrid_files[file_name]
                    transcription_path = os.path.join(root, tg_name)
                if wav_path is None and not self.parse_text_only_files:
                    self.transcriptions_without_wavs.append(transcription_path)
                    continue
                if transcription_path is None:
                    self.no_transcription_files.append(wav_path)

                try:
                    file = parse_file(
                        file_name,
                        wav_path,
                        transcription_path,
                        relative_path,
                        self.speaker_characters,
                        self.sample_rate,
                        self.punctuation,
                        self.clitic_markers,
                    )
                    self.add_file(file)
                except TextParseError as e:
                    self.decode_error_files.append(e)
                except TextGridParseError as e:
                    self.textgrid_read_errors[e.file_name] = e
        if self.decode_error_files or self.textgrid_read_errors:
            self.logger.info(
                "There were some issues with files in the corpus. "
                "Please look at the log file or run the validator for more information."
            )
            if self.decode_error_files:
                self.logger.debug(
                    f"There were {len(self.decode_error_files)} errors decoding text files:"
                )
                self.logger.debug(", ".join(self.decode_error_files))
            if self.textgrid_read_errors:
                self.logger.debug(
                    f"There were {len(self.textgrid_read_errors)} errors decoding reading TextGrid files:"
                )
                for f, e in self.textgrid_read_errors.items():
                    self.logger.debug(f"{f}: {e.error}")

        self.logger.debug(f"Parsed corpus directory in {time.time()-begin_time} seconds")

    def add_file(self, file: File) -> None:
        """
        Add a file to the corpus

        Parameters
        ----------
        file: :class:`~montreal_forced_aligner.corpus.classes.File`
            File to be added
        """
        self.files[file.name] = file
        for speaker in file.speaker_ordering:
            if speaker.name not in self.speakers:
                self.speakers[speaker.name] = speaker
            else:
                self.speakers[speaker.name].merge(speaker)
        for u in file.utterances.values():
            self.utterances[u.name] = u
            if u.text:
                self.word_counts.update(u.text.split())

    def get_word_frequency(self, dictionary: DictionaryType) -> Dict[str, float]:
        """
        Calculate the word frequency across all the texts in the corpus

        Parameters
        ----------
        dictionary: :class:`~montreal_forced_aligner.dictionary.Dictionary`
            Dictionary to use for looking up subwords

        Returns
        -------
        Dict[str, float]
            Dictionary of words and their relative frequencies
        """
        word_counts = Counter()
        for u in self.utterances.values():
            text = u.text
            speaker = u.speaker
            d = dictionary.get_dictionary(speaker)
            new_text = []
            text = text.split()
            for t in text:

                lookup = d.split_clitics(t)
                if lookup is None:
                    continue
                new_text.extend(x for x in lookup if x != "")
            word_counts.update(new_text)
        return {k: v / sum(word_counts.values()) for k, v in word_counts.items()}

    @property
    def word_set(self) -> List[str]:
        """Set of words used in the corpus"""
        return sorted(self.word_counts)

    def add_utterance(self, utterance: Utterance) -> None:
        """
        Add an utterance to the corpus

        Parameters
        ----------
        utterance: :class:`~montreal_forced_aligner.corpus.classes.Utterance`
            Utterance to add
        """
        self.utterances[utterance.name] = utterance
        if utterance.speaker.name not in self.speakers:
            self.speakers[utterance.speaker.name] = utterance.speaker
        if utterance.file.name not in self.files:
            self.files[utterance.file.name] = utterance.file

    def delete_utterance(self, utterance: Union[str, Utterance]) -> None:
        """
        Delete an utterance from the corpus

        Parameters
        ----------
        utterance: :class:`~montreal_forced_aligner.corpus.classes.Utterance`
            Utterance to delete
        """
        if isinstance(utterance, str):
            utterance = self.utterances[utterance]
        utterance.speaker.delete_utterance(utterance)
        utterance.file.delete_utterance(utterance)
        del self.utterances[utterance.name]

    def initialize_jobs(self) -> None:
        """
        Initialize the corpus's Jobs
        """
        if len(self.speakers) < self.num_jobs:
            self.num_jobs = len(self.speakers)
        self.jobs = [Job(i) for i in range(self.num_jobs)]
        job_ind = 0
        for s in self.speakers.values():
            self.jobs[job_ind].add_speaker(s)
            job_ind += 1
            if job_ind == self.num_jobs:
                job_ind = 0

    def initialize_corpus(
        self,
        dictionary: Optional[DictionaryType] = None,
        feature_config: Optional[FeatureConfig] = None,
    ) -> None:
        """
        Initialize corpus for use

        Parameters
        ----------
        dictionary: :class:`~montreal_forced_aligner.dictionary.Dictionary`, optional
            Dictionary to use
        feature_config: :class:`~montreal_forced_aligner.config.feature.FeatureConfig`, optional
            Feature configuration to use
        """
        if not self.files:
            raise CorpusError(
                "There were no wav files found for transcribing this corpus. Please validate the corpus. "
                "This error can also be caused if you're trying to find non-wav files without sox available "
                "on the system path."
            )

        if dictionary is not None:
            for speaker in self.speakers.values():
                speaker.set_dictionary(dictionary.get_dictionary(speaker.name))
        self.initialize_jobs()
        for j in self.jobs:
            j.set_feature_config(feature_config)
        self.feature_config = feature_config
        self.write()
        self.split()
        if self.feature_config is not None:
            try:
                self.generate_features()
            except Exception as e:
                if isinstance(e, KaldiProcessingError):
                    log_kaldi_errors(e.error_logs, self.logger)
                    e.update_log_file(self.logger.handlers[0].baseFilename)
                raise

    @property
    def num_utterances(self) -> int:
        """Number of utterances in the corpus"""
        return len(self.utterances)

    @property
    def features_directory(self) -> str:
        """Feature directory of the corpus"""
        return os.path.join(self.output_directory, "features")

    @property
    def features_log_directory(self) -> str:
        """Feature log directory"""
        return os.path.join(self.split_directory, "log")

    def speaker_utterance_info(self) -> str:
        """
        Construct message for analyzing high level detail about speakers and their utterances

        Returns
        -------
        str
            Analysis string
        """
        num_speakers = len(self.speakers)
        if not num_speakers:
            raise CorpusError(
                "There were no sound files found of the appropriate format. Please double check the corpus path "
                "and/or run the validation utility (mfa validate)."
            )
        average_utterances = sum(len(x.utterances) for x in self.speakers.values()) / num_speakers
        msg = (
            f"Number of speakers in corpus: {num_speakers}, "
            f"average number of utterances per speaker: {average_utterances}"
        )
        return msg

    @property
    def split_directory(self) -> str:
        """Directory used to store information split by job"""
        directory = os.path.join(self.output_directory, f"split{self.num_jobs}")
        return directory

    def generate_features(self, overwrite: bool = False, compute_cmvn: bool = True) -> None:
        """
        Generate features for the corpus

        Parameters
        ----------
        overwrite: bool
            Flag for whether to ignore existing files, defaults to False
        compute_cmvn: bool
            Flag for whether to compute CMVN, defaults to True
        """
        if not overwrite and os.path.exists(os.path.join(self.output_directory, "feats.scp")):
            return
        self.logger.info(f"Generating base features ({self.feature_config.type})...")
        if self.feature_config.type == "mfcc":
            mfcc(self)
        self.combine_feats()
        if compute_cmvn:
            self.logger.info("Calculating CMVN...")
            calc_cmvn(self)
        self.write()
        self.split()

    def compute_vad(self) -> None:
        """
        Compute Voice Activity Dectection features over the corpus
        """
        if os.path.exists(os.path.join(self.split_directory, "vad.0.scp")):
            self.logger.info("VAD already computed, skipping!")
            return
        self.logger.info("Computing VAD...")
        compute_vad(self)

    def combine_feats(self) -> None:
        """
        Combine feature generation results and store relevant information
        """
        split_directory = self.split_directory
        ignore_check = []
        for job in self.jobs:
            feats_paths = job.construct_path_dictionary(split_directory, "feats", "scp")
            lengths_paths = job.construct_path_dictionary(
                split_directory, "utterance_lengths", "scp"
            )
            for dict_name in job.current_dictionary_names:
                path = feats_paths[dict_name]
                lengths_path = lengths_paths[dict_name]
                if os.path.exists(lengths_path):
                    with open(lengths_path, "r") as inf:
                        for line in inf:
                            line = line.strip()
                            utt, length = line.split()
                            length = int(length)
                            if length < 13:  # Minimum length to align one phone plus silence
                                self.utterances[utt].ignored = True
                                ignore_check.append(utt)
                            self.utterances[utt].feature_length = length
                with open(path, "r") as inf:
                    for line in inf:
                        line = line.strip()
                        if line == "":
                            continue
                        f = line.split(maxsplit=1)
                        if self.utterances[f[0]].ignored:
                            continue
                        self.utterances[f[0]].features = f[1]
        for u, utterance in self.utterances.items():
            if utterance.features is None:
                utterance.ignored = True
                ignore_check.append(u)
        if ignore_check:
            self.logger.warning(
                "There were some utterances ignored due to short duration, see the log file for full "
                "details or run `mfa validate` on the corpus."
            )
            self.logger.debug(
                f"The following utterances were too short to run alignment: "
                f"{' ,'.join(ignore_check)}"
            )
        self.write()

    def get_feat_dim(self) -> int:
        """
        Calculate the feature dimension for the corpus

        Returns
        -------
        int
            Dimension of feature vectors
        """
        feature_string = self.jobs[0].construct_base_feature_string(self)
        with open(os.path.join(self.features_log_directory, "feat-to-dim.log"), "w") as log_file:
            dim_proc = subprocess.Popen(
                [thirdparty_binary("feat-to-dim"), feature_string, "-"],
                stdout=subprocess.PIPE,
                stderr=log_file,
            )
            stdout, stderr = dim_proc.communicate()
            feats = stdout.decode("utf8").strip()
        return int(feats)

    def write(self) -> None:
        """
        Output information to the temporary directory for later loading
        """
        self._write_speakers()
        self._write_files()
        self._write_utterances()
        self._write_spk2utt()
        self._write_feats()

    def _write_spk2utt(self):
        """Write spk2utt scp file for Kaldi"""
        data = {
            speaker.name: sorted(speaker.utterances.keys()) for speaker in self.speakers.values()
        }
        output_mapping(data, os.path.join(self.output_directory, "spk2utt.scp"))

    def write_utt2spk(self):
        """Write utt2spk scp file for Kaldi"""
        data = {u.name: u.speaker.name for u in self.utterances.values()}
        output_mapping(data, os.path.join(self.output_directory, "utt2spk.scp"))

    def _write_feats(self):
        """Write feats scp file for Kaldi"""
        if any(x.features is not None for x in self.utterances.values()):
            with open(os.path.join(self.output_directory, "feats.scp"), "w", encoding="utf8") as f:
                for utterance in self.utterances.values():
                    if not utterance.features:
                        continue
                    f.write(f"{utterance.name} {utterance.features}\n")

    def _write_speakers(self):
        """Write speaker information for speeding up future runs"""
        to_save = []
        for speaker in self.speakers.values():
            to_save.append(speaker.meta)
        with open(os.path.join(self.output_directory, "speakers.yaml"), "w", encoding="utf8") as f:
            yaml.safe_dump(to_save, f)

    def _write_files(self):
        """Write file information for speeding up future runs"""
        to_save = []
        for file in self.files.values():
            to_save.append(file.meta)
        with open(os.path.join(self.output_directory, "files.yaml"), "w", encoding="utf8") as f:
            yaml.safe_dump(to_save, f)

    def _write_utterances(self):
        """Write utterance information for speeding up future runs"""
        to_save = []
        for utterance in self.utterances.values():
            to_save.append(utterance.meta)
        with open(
            os.path.join(self.output_directory, "utterances.yaml"), "w", encoding="utf8"
        ) as f:
            yaml.safe_dump(to_save, f)

    def split(self) -> None:
        """Create split directory and output information from Jobs"""
        split_dir = self.split_directory
        os.makedirs(os.path.join(split_dir, "log"), exist_ok=True)
        self.logger.info("Setting up training data...")
        for job in self.jobs:
            job.output_to_directory(split_dir)



"""Class definitions for aligning with pretrained acoustic models"""
from __future__ import annotations

import os
from collections import Counter
from typing import TYPE_CHECKING, Optional

from ..multiprocessing import generate_pronunciations
from .base import BaseAligner

if TYPE_CHECKING:
    from logging import Logger

    from ..config import AlignConfig
    from ..corpus import Corpus
    from ..dictionary import Dictionary
    from ..models import AcousticModel

__all__ = ["PretrainedAligner"]


class PretrainedAligner(BaseAligner):
    """
    Class for aligning a dataset using a pretrained acoustic model

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus object for the dataset
    dictionary : :class:`~montreal_forced_aligner.dictionary.Dictionary`
        Dictionary object for the pronunciation dictionary
    acoustic_model : :class:`~montreal_forced_aligner.models.AcousticModel`
        Archive containing the acoustic model and pronunciation dictionary
    align_config : :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Configuration for alignment
    temp_directory : str, optional
        Specifies the temporary directory root to save files need for Kaldi.
        If not specified, it will be set to ``~/Documents/MFA``
    debug: bool
        Flag for debug mode, default is False
    verbose: bool
        Flag for verbose mode, default is False
    logger: :class:`~logging.Logger`
        Logger to use
    """

    def __init__(
        self,
        corpus: Corpus,
        dictionary: Dictionary,
        acoustic_model: AcousticModel,
        align_config: AlignConfig,
        temp_directory: Optional[str] = None,
        debug: bool = False,
        verbose: bool = False,
        logger: Optional[Logger] = None,
    ):
        super().__init__(
            corpus,
            dictionary,
            align_config,
            temp_directory,
            debug,
            verbose,
            logger,
            acoustic_model=acoustic_model,
        )
        self.data_directory = corpus.split_directory
        log_dir = os.path.join(self.align_directory, "log")
        os.makedirs(log_dir, exist_ok=True)
        self.align_config.logger = self.logger
        self.logger.info("Done with setup!")

    @property
    def model_directory(self) -> str:
        """Model directory"""
        return os.path.join(self.temp_directory, "model")

    def setup(self) -> None:
        """Set up aligner"""
        self.dictionary.nonsil_phones = self.acoustic_model.meta["phones"]
        super(PretrainedAligner, self).setup()
        self.acoustic_model.export_model(self.align_directory)

    @property
    def ali_paths(self):
        """Alignment archive paths"""
        jobs = [x.align_arguments(self) for x in self.corpus.jobs]
        ali_paths = []
        for j in jobs:
            ali_paths.extend(j.ali_paths.values())
        return ali_paths

    def generate_pronunciations(
        self, output_path: str, calculate_silence_probs: bool = False, min_count: int = 1
    ) -> None:
        """
        Generate pronunciation probabilities for the dictionary

        Parameters
        ----------
        output_path: str
            Path to save new dictionary
        calculate_silence_probs: bool
            Flag for whether to calculate silence probabilities, default is False
        min_count: int
            Specifies the minimum count of words to include in derived probabilities, default is 1
        """
        pron_counts, utt_mapping = generate_pronunciations(self)
        if self.dictionary.has_multiple:
            dictionary_mapping = self.dictionary.dictionary_mapping()
        else:
            dictionary_mapping = {self.dictionary.name: self.dictionary}
        for dict_name, dictionary in dictionary_mapping.items():
            counts = pron_counts[dict_name]
            mapping = utt_mapping[dict_name]
            if calculate_silence_probs:
                sil_before_counts = Counter()
                nonsil_before_counts = Counter()
                sil_after_counts = Counter()
                nonsil_after_counts = Counter()
                sils = ["<s>", "</s>", "<eps>"]
                for v in mapping.values():
                    for i, w in enumerate(v):
                        if w in sils:
                            continue
                        prev_w = v[i - 1]
                        next_w = v[i + 1]
                        if prev_w in sils:
                            sil_before_counts[w] += 1
                        else:
                            nonsil_before_counts[w] += 1
                        if next_w in sils:
                            sil_after_counts[w] += 1
                        else:
                            nonsil_after_counts[w] += 1

            dictionary.pronunciation_probabilities = True
            for word, prons in dictionary.words.items():
                if word not in counts:
                    for p in prons:
                        p["probability"] = 1
                else:
                    total = 0
                    best_pron = 0
                    best_count = 0
                    for p in prons:
                        p["probability"] = min_count
                        if p["pronunciation"] in counts[word]:
                            p["probability"] += counts[word][p["pronunciation"]]
                        total += p["probability"]
                        if p["probability"] > best_count:
                            best_pron = p["pronunciation"]
                            best_count = p["probability"]
                    for p in prons:
                        if p["pronunciation"] == best_pron:
                            p["probability"] = 1
                        else:
                            p["probability"] /= total
                    dictionary.words[word] = prons
            dictionary.export_lexicon(output_path, probability=True)



"""Multiprocessing functions for generating features"""
from __future__ import annotations

import os
import shutil
import subprocess
from typing import TYPE_CHECKING, Dict, List, Union

from ..helper import load_scp, make_safe
from ..multiprocessing import run_mp, run_non_mp
from ..utils import thirdparty_binary

if TYPE_CHECKING:
    SpeakerCharacterType = Union[str, int]
    from ..corpus import Corpus
    from ..corpus.classes import ConfigDict

__all__ = ["mfcc", "compute_vad", "calc_cmvn"]


def mfcc_func(
    log_path: str,
    dictionaries: List[str],
    feats_scp_paths: Dict[str, str],
    lengths_paths: Dict[str, str],
    segment_paths: Dict[str, str],
    wav_paths: Dict[str, str],
    mfcc_options: ConfigDict,
) -> None:
    """
    Multiprocessing function for generating MFCC features

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feats_scp_paths: Dict[str, str]
        Dictionary of feature scp files per dictionary name
    lengths_paths: Dict[str, str]
        Dictionary of feature lengths files per dictionary name
    segment_paths: Dict[str, str]
        Dictionary of segment scp files per dictionary name
    wav_paths: Dict[str, str]
        Dictionary of sound file scp files per dictionary name
    mfcc_options: ConfigDict
        Options for MFCC generation
    """
    with open(log_path, "w") as log_file:
        for dict_name in dictionaries:
            mfcc_base_command = [thirdparty_binary("compute-mfcc-feats"), "--verbose=2"]
            raw_ark_path = feats_scp_paths[dict_name].replace(".scp", ".ark")
            for k, v in mfcc_options.items():
                mfcc_base_command.append(f"--{k.replace('_', '-')}={make_safe(v)}")
            if os.path.exists(segment_paths[dict_name]):
                mfcc_base_command += ["ark:-", "ark:-"]
                seg_proc = subprocess.Popen(
                    [
                        thirdparty_binary("extract-segments"),
                        f"scp,p:{wav_paths[dict_name]}",
                        segment_paths[dict_name],
                        "ark:-",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    env=os.environ,
                )
                comp_proc = subprocess.Popen(
                    mfcc_base_command,
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    stdin=seg_proc.stdout,
                    env=os.environ,
                )
            else:
                mfcc_base_command += [f"scp,p:{wav_paths[dict_name]}", "ark:-"]
                comp_proc = subprocess.Popen(
                    mfcc_base_command, stdout=subprocess.PIPE, stderr=log_file, env=os.environ
                )
            copy_proc = subprocess.Popen(
                [
                    thirdparty_binary("copy-feats"),
                    "--compress=true",
                    "ark:-",
                    f"ark,scp:{raw_ark_path},{feats_scp_paths[dict_name]}",
                ],
                stdin=comp_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            copy_proc.communicate()

            utt_lengths_proc = subprocess.Popen(
                [
                    thirdparty_binary("feat-to-len"),
                    f"scp:{feats_scp_paths[dict_name]}",
                    f"ark,t:{lengths_paths[dict_name]}",
                ],
                stderr=log_file,
                env=os.environ,
            )
            utt_lengths_proc.communicate()


def mfcc(corpus: Corpus) -> None:
    """
    Multiprocessing function that converts sound files into MFCCs

    See http://kaldi-asr.org/doc/feat.html and
    http://kaldi-asr.org/doc/compute-mfcc-feats_8cc.html for more details on how
    MFCCs are computed.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/make_mfcc.sh
    for the bash script this function was based on.

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus to generate MFCC features for
    """
    log_directory = os.path.join(corpus.split_directory, "log")
    os.makedirs(log_directory, exist_ok=True)

    jobs = [job.mfcc_arguments(corpus) for job in corpus.jobs]
    if corpus.use_mp:
        run_mp(mfcc_func, jobs, log_directory)
    else:
        run_non_mp(mfcc_func, jobs, log_directory)


def calc_cmvn(corpus: Corpus) -> None:
    """
    Calculate CMVN statistics for speakers

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus to run CMVN calculation
    """
    spk2utt = os.path.join(corpus.output_directory, "spk2utt.scp")
    feats = os.path.join(corpus.output_directory, "feats.scp")
    cmvn_directory = os.path.join(corpus.features_directory, "cmvn")
    os.makedirs(cmvn_directory, exist_ok=True)
    cmvn_ark = os.path.join(cmvn_directory, "cmvn.ark")
    cmvn_scp = os.path.join(cmvn_directory, "cmvn.scp")
    log_path = os.path.join(cmvn_directory, "cmvn.log")
    with open(log_path, "w") as logf:
        subprocess.call(
            [
                thirdparty_binary("compute-cmvn-stats"),
                f"--spk2utt=ark:{spk2utt}",
                f"scp:{feats}",
                f"ark,scp:{cmvn_ark},{cmvn_scp}",
            ],
            stderr=logf,
            env=os.environ,
        )
    shutil.copy(cmvn_scp, os.path.join(corpus.output_directory, "cmvn.scp"))
    for s, cmvn in load_scp(cmvn_scp).items():
        corpus.speakers[s].cmvn = cmvn
    corpus.split()


def compute_vad_func(
    log_path: str,
    dictionaries: List[str],
    feats_scp_paths: Dict[str, str],
    vad_scp_paths: Dict[str, str],
    vad_options: ConfigDict,
) -> None:
    """
    Multiprocessing function to compute voice activity detection

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feats_scp_paths: Dict[str, str]
        Dictionary of feature scp files per dictionary name
    vad_scp_paths: Dict[str, str]
        Dictionary of vad scp files per dictionary name
    vad_options: ConfigDict
        Options for VAD
    """
    with open(log_path, "w") as log_file:
        for dict_name in dictionaries:
            feats_scp_path = feats_scp_paths[dict_name]
            vad_scp_path = vad_scp_paths[dict_name]
            vad_proc = subprocess.Popen(
                [
                    thirdparty_binary("compute-vad"),
                    f"--vad-energy-mean-scale={vad_options['energy_mean_scale']}",
                    f"--vad-energy-threshold={vad_options['energy_threshold']}",
                    "scp:" + feats_scp_path,
                    f"ark,t:{vad_scp_path}",
                ],
                stderr=log_file,
                env=os.environ,
            )
            vad_proc.communicate()


def compute_vad(corpus: Corpus) -> None:
    """
    Compute VAD for a corpus

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus to compute VAD
    """
    log_directory = os.path.join(corpus.split_directory, "log")
    os.makedirs(log_directory, exist_ok=True)
    jobs = [x.vad_arguments(corpus) for x in corpus.jobs]
    if corpus.use_mp:
        run_mp(compute_vad_func, jobs, log_directory)
    else:
        run_non_mp(compute_vad_func, jobs, log_directory)



"""Class definitions for Segmentation based on voice activity in MFA"""
from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING, Dict, List, Optional

from .config import TEMP_DIR
from .exceptions import KaldiProcessingError
from .multiprocessing.ivector import segment_vad
from .utils import log_kaldi_errors, parse_logs

if TYPE_CHECKING:
    from logging import Logger

    from .config import ConfigDict, SegmentationConfig
    from .corpus import Corpus

SegmentationType = List[Dict[str, float]]

__all__ = ["Segmenter"]


class Segmenter:
    """
    Class for performing speaker classification

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.TranscribeCorpus`
        Corpus object for the dataset
    segmentation_config : :class:`~montreal_forced_aligner.config.SegmentationConfig`
        Configuration for alignment
    temp_directory : str, optional
        Specifies the temporary directory root to save files need for Kaldi.
        If not specified, it will be set to ``~/Documents/MFA``
    debug : bool
        Flag for running in debug mode, defaults to false
    verbose : bool
        Flag for running in verbose mode, defaults to false
    logger : :class:`~logging.Logger`, optional
        Logger to use
    """

    def __init__(
        self,
        corpus: Corpus,
        segmentation_config: SegmentationConfig,
        temp_directory: Optional[str] = None,
        debug: Optional[bool] = False,
        verbose: Optional[bool] = False,
        logger: Optional[Logger] = None,
    ):
        self.corpus = corpus
        self.segmentation_config = segmentation_config

        if not temp_directory:
            temp_directory = TEMP_DIR
        self.temp_directory = temp_directory
        self.debug = debug
        self.verbose = verbose
        self.logger = logger
        self.uses_cmvn = False
        self.uses_slices = False
        self.uses_vad = False
        self.speaker_independent = True
        self.setup()

    @property
    def segmenter_directory(self) -> str:
        """Temporary directory for segmentation"""
        return os.path.join(self.temp_directory, "segmentation")

    @property
    def vad_options(self) -> ConfigDict:
        """Options for performing VAD"""
        return {
            "energy_threshold": self.segmentation_config.energy_threshold,
            "energy_mean_scale": self.segmentation_config.energy_mean_scale,
        }

    @property
    def use_mp(self) -> bool:
        """Flag for whether to use multiprocessing"""
        return self.segmentation_config.use_mp

    def setup(self) -> None:
        """
        Sets up the corpus and segmenter for performing VAD

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        done_path = os.path.join(self.segmenter_directory, "done")
        if os.path.exists(done_path):
            self.logger.info("Classification already done, skipping initialization.")
            return
        dirty_path = os.path.join(self.segmenter_directory, "dirty")
        if os.path.exists(dirty_path):
            shutil.rmtree(self.segmenter_directory)
        log_dir = os.path.join(self.segmenter_directory, "log")
        os.makedirs(log_dir, exist_ok=True)
        try:
            self.corpus.initialize_corpus(None, self.segmentation_config.feature_config)
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise

    def segment(self) -> None:
        """
        Performs VAD and segmentation into utterances

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        log_directory = os.path.join(self.segmenter_directory, "log")
        dirty_path = os.path.join(self.segmenter_directory, "dirty")
        done_path = os.path.join(self.segmenter_directory, "done")
        if os.path.exists(done_path):
            self.logger.info("Classification already done, skipping.")
            return
        try:
            self.corpus.compute_vad()
            self.uses_vad = True
            segment_vad(self)
            parse_logs(log_directory)
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        with open(done_path, "w"):
            pass

    def export_segments(self, output_directory: str) -> None:
        """
        Export the results of segmentation as TextGrids

        Parameters
        ----------
        output_directory: str
            Directory to save segmentation TextGrids
        """
        backup_output_directory = None
        if not self.segmentation_config.overwrite:
            backup_output_directory = os.path.join(self.segmenter_directory, "transcriptions")
            os.makedirs(backup_output_directory, exist_ok=True)
        for f in self.corpus.files.values():
            f.save(output_directory, backup_output_directory)



"""Class definitions for base configuration"""
from __future__ import annotations

from typing import TYPE_CHECKING, Collection

import yaml

from ..exceptions import ConfigError

if TYPE_CHECKING:
    from argparse import Namespace

DEFAULT_PUNCTUATION = r'、。।，@<>"(),.:;¿?¡!\\&%#*~【】，…‥「」『』〝〟″⟨⟩♪・‹›«»～′$+=‘'

DEFAULT_CLITIC_MARKERS = "'’"
DEFAULT_COMPOUND_MARKERS = "-/"
DEFAULT_STRIP_DIACRITICS = ["ː", "ˑ", "̩", "̆", "̑", "̯", "͡", "‿", "͜"]
DEFAULT_DIGRAPHS = ["[dt][szʒʃʐʑʂɕç]", "[aoɔe][ʊɪ]"]


PARSING_KEYS = [
    "punctuation",
    "clitic_markers",
    "compound_markers",
    "multilingual_ipa",
    "strip_diacritics",
    "digraphs",
]

__all__ = ["BaseConfig"]


class BaseConfig:
    """
    Base configuration class
    """

    def update(self, data: dict) -> None:
        """Update configuration parameters"""
        for k, v in data.items():
            if not hasattr(self, k):
                raise ConfigError(f"No field found for key {k}")
            setattr(self, k, v)

    def update_from_args(self, args: Namespace) -> None:
        """Update from command line arguments"""
        if args is not None:
            try:
                self.use_mp = not args.disable_mp
            except AttributeError:
                pass
            try:
                self.debug = args.debug
            except AttributeError:
                pass
            try:
                self.overwrite = args.overwrite
            except AttributeError:
                pass
            try:
                self.cleanup_textgrids = not args.disable_textgrid_cleanup
            except AttributeError:
                pass

    def params(self) -> dict:
        """Configuration parameters"""
        return {}

    def update_from_unknown_args(self, args: Collection[str]) -> None:
        """Update from unknown command line arguments"""
        for i, a in enumerate(args):
            if not a.startswith("--"):
                continue
            name = a.replace("--", "")
            try:
                original_value = getattr(self, name)
            except AttributeError:
                continue
            if not isinstance(original_value, (bool, int, float, str)):
                continue
            try:
                if isinstance(original_value, bool):
                    if args[i + 1].lower() == "true":
                        val = True
                    elif args[i + 1].lower() == "false":
                        val = False
                    elif not original_value:
                        val = True
                    else:
                        continue
                else:
                    val = type(original_value)(args[i + 1])
            except (ValueError):
                continue
            except (IndexError):
                if isinstance(original_value, bool):
                    if not original_value:
                        val = True
                    else:
                        continue
                else:
                    continue
            setattr(self, name, val)

    def save(self, path: str) -> None:
        """
        Dump configuration to path

        Parameters
        ----------
        path: str
            Path to export to
        """
        with open(path, "w", encoding="utf8") as f:
            yaml.dump(self.params(), f)



"""Class definitions for adapting acoustic models"""
from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING, Optional

from ..exceptions import KaldiProcessingError
from ..models import AcousticModel
from ..multiprocessing import (
    align,
    calc_fmllr,
    compile_information,
    compile_train_graphs,
    train_map,
)
from ..utils import log_kaldi_errors
from .base import BaseAligner

if TYPE_CHECKING:
    from logging import Logger

    from ..config import AlignConfig
    from ..corpus import Corpus
    from ..dictionary import Dictionary
    from ..models import MetaDict
    from .pretrained import PretrainedAligner


__all__ = ["AdaptingAligner"]


class AdaptingAligner(BaseAligner):
    """
    Aligner adapts another acoustic model to the current data

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus object for the dataset
    dictionary : :class:`~montreal_forced_aligner.dictionary.Dictionary`
        Dictionary object for the pronunciation dictionary
    pretrained_aligner: :class:`~montreal_forced_aligner.aligner.pretrained.PretrainedAligner`
        Pretrained aligner to use as input to training
    align_config : :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Configuration for alignment
    temp_directory : str, optional
        Specifies the temporary directory root to save files need for Kaldi.
        If not specified, it will be set to ``~/Documents/MFA``
    debug: bool
        Flag for debug mode, default is False
    verbose: bool
        Flag for verbose mode, default is False
    logger: :class:`~logging.Logger`
        Logger to use
    """

    def __init__(
        self,
        corpus: Corpus,
        dictionary: Dictionary,
        previous_aligner: PretrainedAligner,
        align_config: AlignConfig,
        temp_directory: Optional[str] = None,
        debug: bool = False,
        verbose: bool = False,
        logger: Optional[Logger] = None,
    ):
        self.previous_aligner = previous_aligner
        super().__init__(
            corpus,
            dictionary,
            align_config,
            temp_directory,
            debug,
            verbose,
            logger,
            acoustic_model=self.previous_aligner.acoustic_model,
        )
        self.align_config.data_directory = corpus.split_directory
        log_dir = os.path.join(self.align_directory, "log")
        os.makedirs(log_dir, exist_ok=True)
        self.align_config.logger = self.logger
        self.logger.info("Done with setup!")
        self.training_complete = False
        self.mapping_tau = 20

    def setup(self) -> None:
        """Set up the aligner"""
        super().setup()
        self.previous_aligner.align()
        self.acoustic_model.export_model(self.adapt_directory)
        for f in ["final.mdl", "final.alimdl"]:
            p = os.path.join(self.adapt_directory, f)
            if not os.path.exists(p):
                continue
            os.rename(p, os.path.join(self.adapt_directory, f.replace("final", "0")))

    @property
    def align_directory(self) -> str:
        """Align directory"""
        return os.path.join(self.temp_directory, "adapted_align")

    @property
    def adapt_directory(self) -> str:
        """Adapt directory"""
        return os.path.join(self.temp_directory, "adapt")

    @property
    def working_directory(self) -> str:
        """Current working directory"""
        if self.training_complete:
            return self.align_directory
        return self.adapt_directory

    @property
    def working_log_directory(self) -> str:
        """Current log directory"""
        return os.path.join(self.working_directory, "log")

    @property
    def current_model_path(self):
        """Current acoustic model path"""
        if self.training_complete:
            return os.path.join(self.working_directory, "final.mdl")
        return os.path.join(self.working_directory, "0.mdl")

    @property
    def next_model_path(self):
        """Next iteration's acoustic model path"""
        return os.path.join(self.working_directory, "final.mdl")

    def adapt(self) -> None:
        """Run the adaptation"""
        done_path = os.path.join(self.adapt_directory, "done")
        dirty_path = os.path.join(self.adapt_directory, "dirty")
        if os.path.exists(done_path):
            self.logger.info("Adapting already done, skipping.")
            return
        try:
            self.logger.info("Adapting pretrained model...")
            train_map(self)
            self.training_complete = True
            shutil.copyfile(
                os.path.join(self.adapt_directory, "final.mdl"),
                os.path.join(self.align_directory, "final.mdl"),
            )
            shutil.copyfile(
                os.path.join(self.adapt_directory, "final.occs"),
                os.path.join(self.align_directory, "final.occs"),
            )
            shutil.copyfile(
                os.path.join(self.adapt_directory, "tree"),
                os.path.join(self.align_directory, "tree"),
            )
            if os.path.exists(os.path.join(self.adapt_directory, "final.alimdl")):
                shutil.copyfile(
                    os.path.join(self.adapt_directory, "final.alimdl"),
                    os.path.join(self.align_directory, "final.alimdl"),
                )
            if os.path.exists(os.path.join(self.adapt_directory, "lda.mat")):
                shutil.copyfile(
                    os.path.join(self.adapt_directory, "lda.mat"),
                    os.path.join(self.align_directory, "lda.mat"),
                )
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        with open(done_path, "w"):
            pass

    @property
    def meta(self) -> MetaDict:
        """Acoustic model metadata"""
        from datetime import datetime

        from ..utils import get_mfa_version

        data = {
            "phones": sorted(self.dictionary.nonsil_phones),
            "version": get_mfa_version(),
            "architecture": self.acoustic_model.meta["architecture"],
            "train_date": str(datetime.now()),
            "features": self.previous_aligner.align_config.feature_config.params(),
            "multilingual_ipa": self.dictionary.multilingual_ipa,
        }
        if self.dictionary.multilingual_ipa:
            data["strip_diacritics"] = self.dictionary.strip_diacritics
            data["digraphs"] = self.dictionary.digraphs
        return data

    def save(self, path, root_directory=None) -> None:
        """
        Output an acoustic model and dictionary to the specified path

        Parameters
        ----------
        path : str
            Path to save acoustic model and dictionary
        root_directory : str or None
            Path for root directory of temporary files
        """
        directory, filename = os.path.split(path)
        basename, _ = os.path.splitext(filename)
        acoustic_model = AcousticModel.empty(basename, root_directory=root_directory)
        acoustic_model.add_meta_file(self)
        acoustic_model.add_model(self.align_directory)
        if directory:
            os.makedirs(directory, exist_ok=True)
        basename, _ = os.path.splitext(path)
        acoustic_model.dump(path)

    def align(self, subset: Optional[int] = None) -> None:
        """
        Align using the adapted model

        Parameters
        ----------
        subset: int, optional
            Number of utterances to align in corpus
        """
        done_path = os.path.join(self.align_directory, "done")
        dirty_path = os.path.join(self.align_directory, "dirty")
        if os.path.exists(done_path):
            self.logger.info("Alignment already done, skipping.")
            return
        try:
            log_dir = os.path.join(self.align_directory, "log")
            os.makedirs(log_dir, exist_ok=True)
            compile_train_graphs(self)

            self.logger.info("Performing first-pass alignment...")
            self.speaker_independent = True
            align(self)
            unaligned, average_log_like = compile_information(self)
            self.logger.debug(
                f"Prior to SAT, average per frame likelihood (this might not actually mean anything): {average_log_like}"
            )
            if (
                not self.align_config.disable_sat
                and self.previous_aligner.acoustic_model.feature_config.fmllr
                and not os.path.exists(os.path.join(self.align_directory, "trans.0"))
            ):
                self.logger.info("Calculating fMLLR for speaker adaptation...")
                calc_fmllr(self)

                self.speaker_independent = False
                self.logger.info("Performing second-pass alignment...")
                align(self)

                unaligned, average_log_like = compile_information(self)
                self.logger.debug(
                    f"Following SAT, average per frame likelihood (this might not actually mean anything): {average_log_like}"
                )

        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        with open(done_path, "w"):
            pass



"""Class definition for IvectorExtractorTrainer"""
from __future__ import annotations

import os
import shutil
import subprocess
import time
from typing import TYPE_CHECKING, Any, Dict, Optional

from tqdm import tqdm

from ..exceptions import KaldiProcessingError
from ..helper import load_scp
from ..models import IvectorExtractor
from ..multiprocessing.ivector import (
    acc_global_stats,
    acc_ivector_stats,
    extract_ivectors,
    gauss_to_post,
    gmm_gselect,
)
from ..utils import log_kaldi_errors, parse_logs, thirdparty_binary
from .base import BaseTrainer

if TYPE_CHECKING:
    from ..aligner import PretrainedAligner
    from ..config import FeatureConfig
    from ..corpus import Corpus
    from ..dictionary import DictionaryType
    from ..models import MetaDict


IvectorConfigType = Dict[str, Any]


__all__ = ["IvectorExtractorTrainer"]


class IvectorExtractorTrainer(BaseTrainer):
    """
    Trainer for IvectorExtractor

    Attributes
    ----------
    ivector_dimension : int
        Dimension of the extracted job_name-vector
    ivector_period : int
        Number of frames between job_name-vector extractions
    num_iterations : int
        Number of training iterations to perform
    num_gselect : int
        Gaussian-selection using diagonal model: number of Gaussians to select
    posterior_scale : float
        Scale on the acoustic posteriors, intended to account for inter-frame correlations
    min_post : float
        Minimum posterior to use (posteriors below this are pruned out)
    subsample : int
        Speeds up training; training on every job_name'th feature
    max_count : int
        The use of this option (e.g. --max-count 100) can make iVectors more consistent for different lengths of utterance, by scaling up the prior term when the data-count exceeds this value. The data-count is after posterior-scaling, so assuming the posterior-scale is 0.1, --max-count 100 starts having effect after 1000 frames, or 10 seconds of data.
    """

    def __init__(self, default_feature_config: FeatureConfig):
        super(IvectorExtractorTrainer, self).__init__(default_feature_config)

        self.ubm_num_iterations = 4
        self.ubm_num_gselect = 30
        self.ubm_num_frames = 500000
        self.ubm_num_gaussians = 256
        self.ubm_num_iterations_init = 20
        self.ubm_initial_gaussian_proportion = 0.5
        self.ubm_min_gaussian_weight = 0.0001

        self.ubm_remove_low_count_gaussians = True

        self.ivector_dimension = 128
        self.num_iterations = 10
        self.num_gselect = 20
        self.posterior_scale = 1.0
        self.silence_weight = 0.0
        self.min_post = 0.025
        self.gaussian_min_count = 100
        self.subsample = 5
        self.max_count = 100
        self.apply_cmn = True
        self.previous_align_directory = None
        self.dubm_training_complete = False
        self.ubm_training_complete = False

    @property
    def meta(self) -> MetaDict:
        """Metadata information for IvectorExtractor"""
        from ..utils import get_mfa_version

        return {
            "version": get_mfa_version(),
            "ivector_dimension": self.ivector_dimension,
            "apply_cmn": self.apply_cmn,
            "num_gselect": self.num_gselect,
            "min_post": self.min_post,
            "posterior_scale": self.posterior_scale,
            "features": self.feature_config.params(),
        }

    @property
    def train_type(self) -> str:
        """Training identifier"""
        return "ivector"

    @property
    def align_directory(self) -> str:
        """Alignment directory"""
        return self.train_directory

    @property
    def ivector_options(self) -> MetaDict:
        """Options for ivector training and extracting"""
        return {
            "subsample": self.subsample,
            "num_gselect": self.num_gselect,
            "posterior_scale": self.posterior_scale,
            "min_post": self.min_post,
            "silence_weight": self.silence_weight,
            "max_count": self.max_count,
            "ivector_dimension": self.ivector_dimension,
            "sil_phones": self.dictionary.silence_csl,
        }

    @property
    def current_ie_path(self) -> str:
        """Current ivector extractor model path"""
        if (
            self.training_complete
            or self.iteration is None
            or self.iteration > self.num_iterations
        ):
            return os.path.join(self.working_directory, "final.ie")
        return os.path.join(self.working_directory, f"{self.iteration}.ie")

    @property
    def next_ie_path(self) -> str:
        """Next iteration's ivector extractor model path"""
        if self.iteration > self.num_iterations:
            return os.path.join(self.working_directory, "final.ie")
        return os.path.join(self.working_directory, f"{self.iteration + 1}.ie")

    @property
    def dubm_path(self) -> str:
        """DUBM model path"""
        return os.path.join(self.working_directory, "final.dubm")

    @property
    def current_dubm_path(self) -> str:
        """Current iteration's DUBM model path"""
        if self.dubm_training_complete:
            return os.path.join(self.working_directory, "final.dubm")
        return os.path.join(self.working_directory, f"{self.iteration}.dubm")

    @property
    def next_dubm_path(self) -> str:
        """Next iteration's DUBM model path"""
        if self.dubm_training_complete:
            return os.path.join(self.working_directory, "final.dubm")
        return os.path.join(self.working_directory, f"{self.iteration + 1}.dubm")

    @property
    def ie_path(self) -> str:
        """Ivector extractor model path"""
        return os.path.join(self.working_directory, "final.ie")

    @property
    def model_path(self) -> str:
        """Acoustic model path"""
        return os.path.join(self.working_directory, "final.mdl")

    def train_ubm_iteration(self) -> None:
        """
        Run an iteration of UBM training
        """
        # Accumulate stats
        acc_global_stats(self)
        self.iteration += 1

    def finalize_train_ubm(self) -> None:
        """Finalize DUBM training"""
        final_dubm_path = os.path.join(self.train_directory, "final.dubm")
        shutil.copy(
            os.path.join(self.train_directory, f"{self.ubm_num_iterations}.dubm"), final_dubm_path
        )
        self.iteration = 0
        self.dubm_training_complete = True

    def train_ubm(self) -> None:
        """
        Train UBM for ivector extractor

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        # train diag ubm
        dirty_path = os.path.join(self.train_directory, "dirty")
        final_ubm_path = os.path.join(self.train_directory, "final.ubm")
        if os.path.exists(final_ubm_path):
            return
        try:
            begin = time.time()
            self.logger.info("Initializing diagonal UBM...")
            # Initialize model from E-M in memory
            log_directory = os.path.join(self.train_directory, "log")
            num_gauss_init = int(
                self.ubm_initial_gaussian_proportion * int(self.ubm_num_gaussians)
            )
            log_path = os.path.join(log_directory, "gmm_init.log")
            feat_name = self.feature_config.feature_id
            all_feats_path = os.path.join(self.corpus.output_directory, f"{feat_name}.scp")
            feature_string = self.corpus.jobs[0].construct_base_feature_string(
                self.corpus, all_feats=True
            )
            with open(all_feats_path, "w") as outf:
                for i in self.corpus.jobs:
                    feat_paths = i.construct_path_dictionary(self.data_directory, "feats", "scp")
                    for p in feat_paths.values():
                        with open(p) as inf:
                            for line in inf:
                                outf.write(line)
            self.iteration = 0
            with open(log_path, "w") as log_file:
                gmm_init_proc = subprocess.Popen(
                    [
                        thirdparty_binary("gmm-global-init-from-feats"),
                        f"--num-threads={self.corpus.num_jobs}",
                        f"--num-frames={self.ubm_num_frames}",
                        f"--num_gauss={self.ubm_num_gaussians}",
                        f"--num_gauss_init={num_gauss_init}",
                        f"--num_iters={self.ubm_num_iterations_init}",
                        feature_string,
                        self.current_dubm_path,
                    ],
                    stderr=log_file,
                )
                gmm_init_proc.communicate()
            # Store Gaussian selection indices on disk
            gmm_gselect(self)
            final_dubm_path = os.path.join(self.train_directory, "final.dubm")

            if not os.path.exists(final_dubm_path):
                self.logger.info("Training diagonal UBM...")
                with tqdm(total=self.ubm_num_iterations) as pbar:
                    while self.iteration < self.ubm_num_iterations + 1:
                        self.train_ubm_iteration()
                        pbar.update(1)
            self.finalize_train_ubm()
            parse_logs(log_directory)
            self.logger.info("Finished training UBM!")
            self.logger.debug(f"UBM training took {time.time() - begin} seconds")
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise

    def init_training(
        self,
        identifier: str,
        temporary_directory: str,
        corpus: Corpus,
        dictionary: DictionaryType,
        previous_trainer: Optional[PretrainedAligner] = None,
    ) -> None:
        """
        Initialize ivector extractor training

        Parameters
        ----------
        identifier: str
            Identifier for the training block
        temporary_directory: str
            Root temporary directory to save
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to use
        dictionary: DictionaryType
            Dictionary to use
        previous_trainer: :class:`~montreal_forced_aligner.trainers.base.BaseTrainer`, optional
            Previous trainer to initialize from
        """
        self._setup_for_init(identifier, temporary_directory, corpus, dictionary, previous_trainer)
        done_path = os.path.join(self.train_directory, "done")
        if os.path.exists(done_path):
            self.logger.info(f"{self.identifier} training already done, skipping initialization.")
            return
        shutil.copyfile(
            previous_trainer.current_model_path, os.path.join(self.train_directory, "final.mdl")
        )
        for p in previous_trainer.ali_paths:
            shutil.copyfile(p, p.replace(previous_trainer.working_directory, self.train_directory))
        self.corpus.write_utt2spk()
        begin = time.time()
        self.previous_align_directory = previous_trainer.align_directory

        self.train_ubm()
        self.init_ivector_train()
        self.logger.info("Initialization complete!")
        self.logger.debug(f"Initialization took {time.time() - begin} seconds")

    def init_ivector_train(self) -> None:
        """
        Initialize ivector extractor training

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        dirty_path = os.path.join(self.train_directory, "dirty")
        try:
            init_ie_path = os.path.join(self.train_directory, "0.ie")
            if os.path.exists(init_ie_path):
                return
            self.iteration = 0
            begin = time.time()
            # Initialize job_name-vector extractor
            log_directory = os.path.join(self.train_directory, "log")
            log_path = os.path.join(log_directory, "init.log")
            diag_ubm_path = os.path.join(self.train_directory, "final.dubm")
            full_ubm_path = os.path.join(self.train_directory, "final.ubm")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [thirdparty_binary("gmm-global-to-fgmm"), diag_ubm_path, full_ubm_path],
                    stderr=log_file,
                )
                subprocess.call(
                    [
                        thirdparty_binary("ivector-extractor-init"),
                        f"--ivector-dim={self.ivector_dimension}",
                        "--use-weights=false",
                        full_ubm_path,
                        self.current_ie_path,
                    ],
                    stderr=log_file,
                )

            # Do Gaussian selection and posterior extraction
            gauss_to_post(self)
            parse_logs(log_directory)
            self.logger.debug(f"Initialization ivectors took {time.time() - begin} seconds")
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise

    def align(self, subset: Optional[int] = None):
        """Overwrite align function to export IvectorExtractor to align directory"""
        self.save(os.path.join(self.align_directory, "ivector_extractor.zip"))

    def training_iteration(self):
        """
        Run an iteration of training
        """
        if os.path.exists(self.next_ie_path):
            self.iteration += 1
            return
        # Accumulate stats and sum
        acc_ivector_stats(self)

        self.iteration += 1

    def finalize_training(self):
        """
        Finalize ivector extractor training
        """
        import numpy as np
        from joblib import dump
        from sklearn.naive_bayes import GaussianNB

        # Rename to final
        shutil.copy(
            os.path.join(self.train_directory, f"{self.num_iterations}.ie"),
            os.path.join(self.train_directory, "final.ie"),
        )
        self.training_complete = True
        self.iteration = None
        extract_ivectors(self)
        x = []
        y = []
        speakers = sorted(self.corpus.speakers.keys())
        for j in self.corpus.jobs:
            arguments = j.extract_ivector_arguments(self)
            for ivector_path in arguments.ivector_paths.values():
                ivec = load_scp(ivector_path)
                for utt, ivector in ivec.items():
                    ivector = [float(x) for x in ivector]
                    s = self.corpus.utterances[utt].speaker.name
                    s_ind = speakers.index(s)
                    y.append(s_ind)
                    x.append(ivector)
        x = np.array(x)
        y = np.array(y)
        clf = GaussianNB()
        clf.fit(x, y)
        clf_param_path = os.path.join(self.train_directory, "speaker_classifier.mdl")
        dump(clf, clf_param_path)
        classes_path = os.path.join(self.train_directory, "speaker_labels.txt")
        with open(classes_path, "w", encoding="utf8") as f:
            for i, s in enumerate(speakers):
                f.write(f"{s} {i}\n")

    def save(self, path: str, root_directory: Optional[str] = None):
        """
        Output IvectorExtractor model

        Parameters
        ----------
        path : str
            Path to save acoustic model and dictionary
        root_directory : str or None
            Path for root directory of temporary files
        """
        directory, filename = os.path.split(path)
        basename, _ = os.path.splitext(filename)
        ivector_extractor = IvectorExtractor.empty(basename, root_directory)
        ivector_extractor.add_meta_file(self)
        ivector_extractor.add_model(self.train_directory)
        os.makedirs(directory, exist_ok=True)
        basename, _ = os.path.splitext(path)
        ivector_extractor.dump(basename)



"""Command line functions for validating corpora"""
from __future__ import annotations

import os
import shutil
import time
from typing import TYPE_CHECKING, List, Optional

from montreal_forced_aligner.command_line.utils import validate_model_arg
from montreal_forced_aligner.config import TEMP_DIR
from montreal_forced_aligner.corpus import Corpus
from montreal_forced_aligner.dictionary import Dictionary
from montreal_forced_aligner.exceptions import ArgumentError
from montreal_forced_aligner.models import AcousticModel
from montreal_forced_aligner.utils import setup_logger
from montreal_forced_aligner.validator import CorpusValidator

if TYPE_CHECKING:
    from argparse import Namespace


__all__ = ["validate_corpus", "validate_args", "run_validate_corpus"]


def validate_corpus(args: Namespace, unknown_args: Optional[List[str]] = None) -> None:
    """
    Run the validation command

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Command line arguments
    unknown_args: List[str]
        Optional arguments that will be passed to configuration objects
    """
    command = "validate"
    all_begin = time.time()
    if not args.temp_directory:
        temp_dir = TEMP_DIR
    else:
        temp_dir = os.path.expanduser(args.temp_directory)
    corpus_name = os.path.basename(args.corpus_directory)
    if corpus_name == "":
        args.corpus_directory = os.path.dirname(args.corpus_directory)
        corpus_name = os.path.basename(args.corpus_directory)
    data_directory = os.path.join(temp_dir, corpus_name)
    shutil.rmtree(data_directory, ignore_errors=True)

    os.makedirs(data_directory, exist_ok=True)
    model_directory = os.path.join(data_directory, "acoustic_models")
    os.makedirs(model_directory, exist_ok=True)
    if getattr(args, "verbose", False):
        log_level = "debug"
    else:
        log_level = "info"
    logger = setup_logger(command, data_directory, console_level=log_level)

    corpus = Corpus(
        args.corpus_directory,
        data_directory,
        speaker_characters=args.speaker_characters,
        num_jobs=getattr(args, "num_jobs", 3),
        logger=logger,
        use_mp=not args.disable_mp,
    )
    if args.acoustic_model_path:
        acoustic_model = AcousticModel(args.acoustic_model_path, root_directory=model_directory)
        acoustic_model.log_details(logger)
        dictionary = Dictionary(
            args.dictionary_path,
            data_directory,
            logger=logger,
            multilingual_ipa=acoustic_model.meta["multilingual_ipa"],
        )
        acoustic_model.validate(dictionary)
    else:
        dictionary = Dictionary(args.dictionary_path, data_directory, logger=logger)

    a = CorpusValidator(
        corpus,
        dictionary,
        temp_directory=data_directory,
        ignore_acoustics=getattr(args, "ignore_acoustics", False),
        test_transcriptions=getattr(args, "test_transcriptions", False),
        use_mp=not args.disable_mp,
        logger=logger,
    )
    begin = time.time()
    a.validate()
    logger.debug(f"Validation took {time.time() - begin} seconds")
    logger.info("All done!")
    logger.debug(f"Done! Everything took {time.time() - all_begin} seconds")
    handlers = logger.handlers[:]
    for handler in handlers:
        handler.close()
        logger.removeHandler(handler)


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    """
    try:
        args.speaker_characters = int(args.speaker_characters)
    except ValueError:
        pass
    if args.test_transcriptions and args.ignore_acoustics:
        raise ArgumentError("Cannot test transcriptions without acoustic feature generation.")
    if not os.path.exists(args.corpus_directory):
        raise (ArgumentError(f"Could not find the corpus directory {args.corpus_directory}."))
    if not os.path.isdir(args.corpus_directory):
        raise (
            ArgumentError(
                f"The specified corpus directory ({args.corpus_directory}) is not a directory."
            )
        )

    args.dictionary_path = validate_model_arg(args.dictionary_path, "dictionary")
    if args.acoustic_model_path:
        args.acoustic_model_path = validate_model_arg(args.acoustic_model_path, "acoustic")


def run_validate_corpus(args: Namespace, unknown: Optional[List[str]] = None) -> None:
    """
    Wrapper function for running corpus validation

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown: List[str]
        Parsed command line arguments to be passed to the configuration objects
    """
    validate_args(args)
    validate_corpus(args, unknown)
